import json
import os

import typer

from statecraft.client import client
from statecraft.user_attributes import UserAttributes
from statecraft.utils import get_default_cache_dir


def setup() -> None:
    cache_dir = get_default_cache_dir()
    username = typer.prompt("Please enter your username")
    email = typer.prompt("Please enter your email address")
    user_attributes = UserAttributes(username=username, email=email, token="")

    # Save user_attributes
    with open(os.path.join(cache_dir, "user_attributes.json"), "w") as json_file:
        json.dump(user_attributes.to_dict(), json_file)

    print("Saved username and email!")

    response_code = client.create_user(username, email)
    if response_code in (200, 201):
        print("User created successfully!")
    else:
        print("User creation failed. Please try again.")



from dataclasses import asdict, dataclass
from typing import Optional


@dataclass
class SSMStateMetadata:
    state_name: str
    prompt: str
    model_name: str
    description: Optional[str]
    keywords: Optional[list[str]] = None

    def to_dict(self):
        return asdict(self)



from datetime import datetime
from typing import Optional

from pydantic import BaseModel


class StateOut(BaseModel):
    state_name: str
    description: Optional[str] = None
    model_name: str
    prompt: str
    keywords: Optional[list[str]] = None
    state_id: int
    user_id: int
    created_at: datetime
    updated_at: datetime


StatesOut = list[StateOut]



"Store, manage and remix states for SSMs and other Stateful models"

from statecraft._setup import setup
from statecraft.core import SSMStateMetadata, StatefulModel, upload_state
from statecraft.states_list import show_available_states



import getpass
import os


def get_default_cache_dir():
    if os.name == "posix":  # For Linux and macOS
        return os.path.expanduser("~/.cache/statecraft/")
    elif os.name == "nt":  # For Windows
        username = getpass.getuser()
        return os.path.expanduser(rf"C:\Users\{username}\.cache\statecraft")
    else:
        raise ValueError(f"Unsupported operating system: {os.name}")



import os
from typing import Optional

from statecraft import StatefulModel
from statecraft.client import client

ALL_DIRS_TO_REMOVE = [".DS_Store"]


def _list_local_states(model_name: str, cache_dir: Optional[str] = None) -> list[str]:
    if cache_dir is None:
        cache_dir = StatefulModel._get_default_cache_dir()

    out = []

    model_dir = os.path.join(cache_dir, model_name, "CURRENT_USER")
    if os.path.isdir(model_dir):
        current_user_states = os.listdir(model_dir)
        out += [dir for dir in current_user_states if dir not in ALL_DIRS_TO_REMOVE]

    other_dirs = os.listdir(os.path.join(cache_dir, model_name))
    other_dirs = [
        dir
        for dir in other_dirs
        if (dir not in ALL_DIRS_TO_REMOVE) and (dir != "CURRENT_USER")
    ]

    for dir in other_dirs:
        if (dir not in ALL_DIRS_TO_REMOVE + ["CURRENT_USER"]) and (os.path.isdir(dir)):
            sub_dirs = os.listdir(os.path.join(cache_dir, model_name, dir))
            out += [
                f"{dir}/{sub_dir}" for sub_dir in sub_dirs if sub_dir not in ALL_DIRS_TO_REMOVE
            ]

    return out


def _list_server_states(model_name: str) -> list[str]:
    return client.get_states(model_name)


def show_available_states(model_name: str, cache_dir: Optional[str] = None) -> None:
    local_states = _list_local_states(model_name, cache_dir)
    server_states = _list_server_states(model_name)

    print(f"Local states for {model_name}:")
    print(local_states)

    print("")
    print(f"Server states for {model_name}:")
    print(server_states)


if __name__ == "__main__":
    states = _list_local_states("state-spaces/mamba-130m-hf")
    print(states)



from pathlib import Path

from setuptools import find_packages, setup

this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

setup(
    name="statecraft",
    version="0.1.1",
    packages=find_packages(exclude=["tests", "docs"]),
    long_description=long_description,
    long_description_content_type="text/markdown",
)



import json
import os
from pathlib import Path
from typing import Union

import requests

from statecraft.metadata import SSMStateMetadata
from statecraft.models import StateOut, StatesOut
from statecraft.user_attributes import UserAttributes
from statecraft.utils import get_default_cache_dir

BASE_URL = "https://www.api.statecrafthub.com/"
LOCAL_URL = "http://localhost/"


class StatecraftClient:

    def __init__(self, base_url: str = BASE_URL):
        self.base_url = base_url
        self.states_url = f"{base_url}states/"

    def get_state(self, model_name: str, state_name: str) -> bytes:
        full_url = os.path.join(self.states_url, model_name, state_name)

        response = requests.get(full_url)

        if response.status_code == 200:
            raw_bytes = response.content
            print("Got state - raw_bytes: ", raw_bytes[:100])
            return raw_bytes
        else:
            raise ValueError(f"Failed to get state: {response.text}")

    def upload_state(
        self,
        metadata: SSMStateMetadata,
        state_path: Union[Path, str],
    ) -> str:
        url = self.states_url

        print(url)

        state_short_name = metadata.state_name.split("/")[-1]

        user_attributes = self._fetch_user_attrs()
        username = user_attributes.username

        query_params = {
            "state_name": f"{username}/{state_short_name}",
            "model_name": metadata.model_name,
            "prompt": metadata.prompt,
        }

        if metadata.description:
            query_params["description"] = metadata.description
        if metadata.keywords:
            query_params["keywords"] = str(metadata.keywords)

        files = {
            "state_file": (
                f"{state_short_name}.pt",
                open(state_path, "rb"),
                "application/octet-stream",
            )
        }

        headers = {
            "accept": "application/json",
        }
        response = requests.post(url, params=query_params, files=files, headers=headers)

        try:
            response.raise_for_status()
        except requests.exceptions.RequestException as e:
            print(f"Error: {e}")

        if response.status_code in (200, 201):
            print(response.text)
            return "State uploaded successfully!"
        else:
            return f"Failed to upload state: {response.status_code}"

    def get_states(self, model_name: str) -> list[str]:
        model_user_name, model_short_name = model_name.split("/")
        query_params = {
            "model_user_name": model_user_name,
            "model_short_name": model_short_name,
        }

        response = requests.get(
            self.states_url,
            params=query_params,
        )
        print(response.text)
        parsed_response: list[dict[str, str]] = response.json()  # StatesOut

        states_list = [
            f'{response_state_dict["model_name"]}/{response_state_dict["state_name"]}'
            for response_state_dict in parsed_response
        ]

        return states_list

    def create_user(self, username: str, email: str) -> int:
        response = requests.post(
            os.path.join(self.base_url, "users"),
            json={"username": username, "email": email},
        )
        response_code = response.status_code
        return response_code

    @classmethod
    def _fetch_user_attrs(cls) -> UserAttributes:
        cache_dir = get_default_cache_dir()

        # Check if user attributes are saved
        if os.path.exists(os.path.join(cache_dir, "user_attributes.json")):
            with open(os.path.join(cache_dir, "user_attributes.json"), "r") as f:
                user_attributes_dict: dict = json.load(f)
                user_attributes = UserAttributes(
                    username=user_attributes_dict["username"],
                    email=user_attributes_dict["email"],
                    token=user_attributes_dict["token"],
                )
                return user_attributes
        else:
            raise ValueError("User attributes not found. Please run statecraft.setup()")


client = StatecraftClient()



import json
import os
from pathlib import Path
from typing import Any, Optional, Union

import torch as t
from transformers import AutoTokenizer, MambaForCausalLM, PreTrainedModel
from transformers.models.mamba.modeling_mamba import MambaCache, MambaCausalLMOutput

from statecraft.client import client
from statecraft.metadata import SSMStateMetadata
from statecraft.utils import get_default_cache_dir


def get_cached_state(
    saved_state_name: Union[Path, str],
    model_name: str,
    state_username: str = "CURRENT_USER",
    cache_dir: Optional[str] = None,
) -> tuple[MambaCache, SSMStateMetadata, str]:
    if cache_dir is None:
        cache_dir = StatefulModel._get_default_cache_dir()

    if "/" in str(saved_state_name):
        base_path = os.path.join(cache_dir, model_name, Path(saved_state_name))
    else:
        base_path = os.path.join(cache_dir, model_name, state_username, Path(saved_state_name))

    is_local = os.path.isdir(base_path)
    if not is_local:
        raise ValueError("Path to saved state must be a directory which exists on the system")

    state = t.load(os.path.join(base_path, "state.pt"))

    with open(os.path.join(base_path, "metadata.json"), "r") as json_file:
        metadata_dict: dict[str, Any] = json.load(json_file)

    for key in ("state_name", "model_name", "prompt"):
        if key not in metadata_dict:
            raise ValueError(f"Metadata must contain key: {key}")

    metadata = SSMStateMetadata(
        state_name=metadata_dict.get("state_name"),  # type: ignore
        model_name=metadata_dict.get("model_name"),  # type: ignore
        prompt=metadata_dict.get("prompt"),  # type: ignore
        description=metadata_dict.get("description"),
        keywords=metadata_dict.get("keywords"),
    )

    print("Metadata: \n", metadata)
    print(f"State loaded from {base_path}")

    return state, metadata, base_path


def upload_state(path: Union[Path, str], model_name: str) -> None:
    _state, metadata, base_path = get_cached_state(path, model_name)

    if metadata.model_name != model_name:
        raise ValueError(
            "Model name in metadata does not match the model name provided.",
            f"Model name in metadata: {metadata.model_name}",
            f"Model name provided: {model_name}",
        )

    print("Uploading state from ", base_path)
    print("Metadata: \n", metadata)

    state_path = os.path.join(base_path, "state.pt")

    # Make API call
    client.upload_state(metadata, state_path)


class StatefulModel(PreTrainedModel):
    def __init__(
        self,
        model: MambaForCausalLM,
        initial_state: Optional[MambaCache],
        device: Optional[str] = None,
        model_name: Optional[str] = None,
    ):
        super().__init__(model.config)
        if initial_state is None:
            initial_state = MambaCache(config=model.config, batch_size=1, device=device)
        self.initial_state: MambaCache = initial_state
        self.model = model
        self.model_name = model_name

    def forward(
        self,
        input_ids: t.Tensor,
        cache_params: Optional[MambaCache] = None,
        reset_sequence_offset: bool = True,
    ) -> MambaCausalLMOutput:
        # TODO: Deal with change in batch size
        if cache_params is None:
            cache_params = self.initial_state
        if reset_sequence_offset:
            cache_params = self._reset_state_offset(cache_params)

        out: MambaCausalLMOutput = self.model(
            input_ids=input_ids, cache_params=cache_params, use_cache=True
        )
        return out

    @classmethod
    def from_pretrained(
        cls,
        model_name: str,
        initial_state_name: Optional[str] = None,
        device: Optional[str] = None,
    ) -> "StatefulModel":
        # Load model from Hugging Face
        model: MambaForCausalLM = MambaForCausalLM.from_pretrained(model_name)  # type: ignore

        # Load initial state
        if initial_state_name is not None:
            initial_state = cls._load_state(model_name, initial_state_name)
        else:
            initial_state = None

        stateful_model = cls(model=model, initial_state=initial_state, model_name=model_name)
        return stateful_model

    def build_state(
        self,
        input_ids: t.Tensor,
        save_state: bool,
        cache_params: Optional[MambaCache] = None,
    ) -> MambaCache:
        out: MambaCausalLMOutput = self.forward(input_ids=input_ids, cache_params=cache_params)
        assert out.cache_params is not None
        if save_state:
            pass
        return out.cache_params

    def rag_generate(self, input_str: str) -> str:
        raise NotImplementedError

    def combine_states(
        self, states: list[MambaCache], weights: Optional[list[float]] = None
    ) -> MambaCache:
        # TODO: Check compatibility

        # Combine states
        if weights is None:
            weights = [1 / len(states)] * len(states)
        raise NotImplementedError

    # STATE SAVING
    @classmethod
    def save_state(
        cls,
        state: MambaCache,
        metadata: SSMStateMetadata,
        path: Optional[Union[Path, str]] = None,
        cache_dir: Optional[str] = None,
    ) -> None:
        if cache_dir is None:
            cache_dir = cls._get_default_cache_dir()
        # Create path to save_location
        if path is None:
            path = metadata.state_name
        base_path = os.path.join(cache_dir, metadata.model_name, "CURRENT_USER", Path(path))

        # Create the cache directory if it doesn't exist
        os.makedirs(base_path, exist_ok=True)

        state_path = os.path.join(base_path, "state.pt")
        metadata_path = os.path.join(base_path, "metadata.json")

        # Save state
        t.save(state, state_path)
        print(f"State saved to {state_path}")

        # Save metadata as json file
        with open(metadata_path, "w") as json_file:
            json.dump(metadata.to_dict(), json_file)
        print(f"Metadata saved to {metadata_path}")

    def save_current_state(self, state_name: str, prompt: str) -> None:
        if self.model_name is None:
            raise ValueError(
                "Method .save_current_state(...) requires the model_name to be set.",
                "This happens automatically when using from_pretrained.",
                "Otherwise you can set it manually or save the state using .save_state(...)",
            )
        metadata = SSMStateMetadata(
            state_name=state_name,
            prompt=prompt,
            model_name=self.model_name,
            description="",
        )
        self.save_state(state=self.initial_state, metadata=metadata)

    # STATE LOADING

    def load_state(self, path: str, cache_dir: Optional[str] = None) -> None:
        if self.model_name is None:
            raise ValueError(
                "Method .load_state(...) requires the model_name to be set.",
                "This happens automatically when using from_pretrained.",
                "Otherwise you can set it manually",
            )
        state = self._load_state(
            model_name=self.model_name, state_name_path=path, cache_dir=cache_dir
        )
        self.update_state(state)

    def update_state(self, state: MambaCache) -> None:
        self.initial_state = state

    def reset_state(self) -> None:
        self.initial_state = MambaCache(config=self.model.config, batch_size=1, device=None)

    # HELPER METHODS

    @classmethod
    def _save_state_binaries(
        cls,
        state_bytes: bytes,
        model_name_path: str,
        state_name_path: str,
        cache_dir: Optional[str] = None,
    ) -> None:
        cache_dir = cls._get_default_cache_dir() if cache_dir is None else cache_dir
        base_path = os.path.join(cache_dir, model_name_path, state_name_path)

        os.makedirs(base_path, exist_ok=False)

        state_path = os.path.join(base_path, "state.pt")
        with open(state_path, "wb") as f:
            f.write(state_bytes)

        metadata_path = os.path.join(base_path, "metadata.json")
        metadata = SSMStateMetadata(
            state_name=state_name_path,
            model_name=model_name_path,
            prompt="See statecrafthub.com for details on the prompt used",
            description=None,
        )

        with open(metadata_path, "w") as json_file:
            json.dump(metadata.to_dict(), json_file)

    @classmethod
    def _load_state(
        cls, model_name: str, state_name_path: str, cache_dir: Optional[str] = None
    ) -> MambaCache:
        try:  # Try loading from local cache
            state, _metadata, _base_path = get_cached_state(
                saved_state_name=state_name_path,
                model_name=model_name,
                cache_dir=cache_dir,
            )
            return state
        except Exception as e:
            pass

        try:  # Try loading from server
            state_bytes = client.get_state(model_name, state_name_path)
        except Exception as e:
            raise ValueError(
                f"Failed to load state from local cache and server.",
                f"Model name: {model_name}",
                f"State name: {state_name_path}",
                e,
            )

        try:
            cls._save_state_binaries(
                state_bytes=state_bytes,
                model_name_path=model_name,
                state_name_path=state_name_path,
                cache_dir=cache_dir,
            )
        except Exception as e:
            raise IOError(
                f"Failed to save state binaries to local cache.",
                f"Model name: {model_name}",
                f"State name: {state_name_path}",
                e,
            )

        state, _metadata, _base_path = get_cached_state(
            saved_state_name=state_name_path,
            model_name=model_name,
            cache_dir=cache_dir,
        )
        return state

    @classmethod
    def _get_default_cache_dir(cls) -> str:
        return get_default_cache_dir()

    def _check_state_compatible(self, state: MambaCache) -> bool:
        raise NotImplementedError

    def _reset_state_offset(self, state: MambaCache) -> MambaCache:
        state.seqlen_offset = 0
        return state
